;; Generated package description from uniquify-files.el  -*- no-byte-compile: t -*-
(define-package "uniquify-files" "1.0.4.0.20220528.40612" "Completion style for files, minimizing directories" '((emacs "25.0")) :commit "1d76b4f0e283afaff2be053d85f8726ffc3abd6e" :url "https://elpa.gnu.org/packages/uniquify-files.html" :authors '(("Stephen Leake" . "stephen_leake@stephe-leake.org")) :maintainer '("Stephen Leake" . "stephen_leake@stephe-leake.org") :keywords '("completion" "table" "uniquify"))
